###########################################
# Basis Selection from Multiple Libraries #
# Updated Dec 18 2009					  #
# Version 1.0-5                           #
###########################################

bsml <- function(y,baseslist,method="bsmlc",maxbas=30,sub.maxbas=c(),backward=NULL,
                 has.control=list(crit="GCV",idf=1.2,ridge=F),
                 bsmlc.approach="gdf",
                 gdf.control=list(nrep=100,tauhat="Rice"),
                 gcvpk.control=list(delta=.1,job=1005,lamlim=c(-6,2),tau=10,ntbl=100,maxtbl=100)
                 ) 
{

    # Let idf=NA if you want idf to be estimated from the data
	
	#### Check if the libraries are entered as a list ####
    if (!is.list(baseslist)) 
      stop("Basis functions must be entered as a list.","\n")   

	#### Check if total number of bases available for selection is greater than maxbas ####
    totalbases <- 0
    starter <- ifelse(is.null(baseslist[[1]]),2,1)
    for (i in starter:length(baseslist)) totalbases <- totalbases+ncol(baseslist[[i]])
    if (totalbases<maxbas) 
      stop("Total number of basis functions in all libraries is less than MAXBAS.","\n")  

    #### Check for columns of all zeroes in "baseslist" ####
    for (i in starter:length(baseslist)) {
      check.zero <- 1:ncol(baseslist[[i]])
      for (j in 1:ncol(baseslist[[i]])) {
          check.zero[j] <- sum(abs(baseslist[[i]][,j]))       
          if (check.zero[j]==0)  
            cat("Column", j, "of Library",i,"contains all zeros.","\n")
      }
      if (min(check.zero)==0) 
        stop("At least one basis function contains all zeros.","\n")     
    }

   #### Select Spatially Adaptive Regression Method ####
   # has = Hybrid Adaptive Splines (with option to estimate IDF)
   # bsmlc = BSML Libraries Combined
   # bsmls = BSML Libraries Separate

   if (method=="bsmls")  { 
       est <- bsmls(y,baseslist,maxbas=maxbas,sub.maxbas=sub.maxbas,backward=backward,
                      gdf.control=gdf.control,
                      gcvpk.control=gcvpk.control)

   }


   if (method=="bsmlc")  {
       est <- bsmlc(y,baseslist,maxbas=maxbas,sub.maxbas=sub.maxbas,backward=backward,
                      approach=bsmlc.approach,
                      gdf.control=gdf.control,
                      gcvpk.control=gcvpk.control) 


   }


   if (method=="has")  {
       est <- has(y,baseslist,maxbas=maxbas,backward=backward,
                    has.control=has.control,
                    gdf.control=gdf.control,
                    gcvpk.control=gcvpk.control) 

 
   }

   est$call=match.call()
   class(est) <- "bsml"
   return(est)

}

print.bsml <- function(x, ...)
{
Call=x$call
Total_Number_of_Bases=x$nb
Chosen_Bases_List=as.matrix(x$chosen.bases.trim)
Generalized_Degree_of_Freedom=x$dof
Estimated_Variance=x$sigma_sq
cat("Call:\n")
print(Call)
cat("\nTotal Number of Bases in Model: (including null bases)\n")
print(Total_Number_of_Bases)
cat("\nChosen Bases List: (excluding null bases)\n")
print(Chosen_Bases_List)
cat("\nGeneralized Degree of Freedom:\n")
print(Generalized_Degree_of_Freedom)
cat("\nEstimated Variance:\n")
print(Estimated_Variance)
}

stdz <- function(x) {
nms <- sqrt(colSums(x^2))
for (i in 1:ncol(x)) {
	x[,i]=x[,i]/nms[i]
}
return(x)
}

###########################
# Hybrid Adaptive Splines #
###########################


has <- function(y,baseslist,maxbas=30,backward=NULL,
                has.control=list(crit="GCV",idf=1.2,ridge=F),
                gdf.control=list(nrep=100,tauhat="Rice"),
                gcvpk.control=list(delta=.1,job=1005,lamlim=c(-6,2),tau=10,ntbl=100,maxtbl=100)) 
{
    nobs <- length(y)
    starter <- ifelse(is.null(baseslist[[1]]),2,1)
	# Check if the number of rows of all libraries are the same
	for (i in starter:length(baseslist)) {
	lib.dim=nrow(as.matrix(baseslist[[i]]))
	if (lib.dim!=nobs) stop("Library ", i, " has the wrong number of rows.", "\n")
	}

	crit=has.control$crit
	idf=has.control$idf
	ridge=has.control$ridge
    coefficients <- coef <- residuals <- fitted.values <- index.bas <- isb <- nb <- NULL
    fits <- y
    nrep <- gdf.control$nrep
    tauhat <- gdf.control$tauhat
    vmu <- 1
    control <- gcvpk.control
    nnull <- ifelse(is.null(baseslist[[1]]),0,ncol(baseslist[[1]]))
    bas <- basel <- NULL
    for (i in 1:length(baseslist)) {
    bas <- cbind(bas,baseslist[[i]])
    }

    # Double-checking if total number of bases is greater than maxbas
    totalbases <- 0
    for (i in starter:length(baseslist)) totalbases <- totalbases+ncol(baseslist[[i]])
    if (totalbases<maxbas) stop("Total number of basis functions in all libraries is less than MAXBAS.","\n")  
	
    # Check if any two columns (bases) among the libraries are the same
    for (i in 2:length(baseslist)) basel <- cbind(basel,baseslist[[i]])
	numcol <- ncol(as.matrix(basel))
	coldist <- 0
	for (j in 1:(numcol-1)) {
		if (var(basel[,j])==0) warning("Column ", j, " is a constant.","\n")
		for (k in (j+1):numcol) {
			coldist <- sum((basel[,j]-basel[,k])^2)
			if (abs(coldist)<1e-20) warning("Column ", j, " and column ", k, " in the combined non-null libraries are the same.","\n","Fit is likely wrong.","\n")
		}
	}
	if (var(basel[,numcol])==0) warning("Column ", numcol, " is a constant.","\n")
	
#    if (min(diag(cov(basel)))>0) colcor <- cor(basel) else stop("Constant function in non-null libraries.","\n")
#    maxcor <- max(colcor-diag(rep(1,ncol(as.matrix(basel)))))
#    if (abs(maxcor-1)<1e-20) stop("Same basis has appeared twice among the libraries.","\n")

    if (is.na(idf)) {
      if (tauhat == "Rice") tauhat <- .5*sqrt(mean(diff(y)^2)/2)
	else if (tauhat == "Robust") {
	  gaps=abs(diff(y))
	  gaps.lb=quantile(gaps,0.025)
	  gaps.ub=quantile(gaps,0.975)
	  gaps=gaps*(1*(gaps>=gaps.lb & gaps<=gaps.ub))
	  tauhat <- .5*sqrt(mean((gaps[gaps>0])^2)/2)
	}
      idf.all <- bf.cost2(y,bas,nnull=nnull,maxbas=maxbas,wkamax=maxbas,nrep=nrep,tauhat=tauhat)$idfhat
	idf <- mean(idf.all)
    }

    nf <- sys.nframe()
    ctl <- eval(control,sys.frame(which=nf))
    if (crit=="GCV") tmpcrit <- 1
    if (crit=="AIC") tmpcrit <- 2
    if (crit=="BIC") tmpcrit <- 3
    if (crit=="Cp")  tmpcrit <- 4
    cond <- 1
    if (is.null(backward) & ridge==T) cond <- 2

    nbas <- ncol(bas)
    if (nbas < maxbas) stop("nbas < maxbas")
    
    check.zero <- 1:nbas
    for (i in 1:nbas) {
        check.zero[i] <- sum(abs(bas[,i]))
        if (check.zero[i]==0)  
            cat("Column ", i, "contains all zeros.","\n")
    }
    if (min(check.zero)==0) 
        stop("At least one column contains all zeros.","\n")               
 
    isb <- coef <- score <- rss <- seq(0, maxbas)
    yy <- hasfit <- adiag <- svals <- prefits <- y 
    hasdev <- hasdf <- nb <- info <- 0
    iwk <- rep(1, nbas)
    iwork <- rep(1,2*maxbas-nnull)
    baswk <- bas 
    wka <- des <- predes <- matrix(0,nobs,maxbas) 
    baswork <- matrix(0,nobs,ncol(bas))
    sigma <- matrix(0,maxbas,maxbas) 
    tbl <- matrix(0,ctl$maxtbl,3)
    lwa <- (maxbas-nnull)*(maxbas-2*nnull+2+nobs)+maxbas+nobs 
    work <- rep(0,lwa)
    fits <- u <- h <- z <- zk <- rep(0,nobs)
    fitswk <- matrix(0,nobs,maxbas)

    hasfit <- .Fortran("has1",
            NOBS = as.integer(nobs),
            Y = as.double(y),
            BAS = as.double(bas),
            NNULL = as.integer(nnull),
            NBAS = as.integer(nbas),
            MAXBAS = as.integer(maxbas),
            CRIT = as.integer(tmpcrit),
            IDF = as.double(idf),
            DELTA = as.double(ctl$delta), 
            COND = as.integer(cond),
            NB = as.integer(nb),
            ISB = as.integer(isb),
            COEF = as.double(coef),
            HASFIT = as.double(hasfit),
            HASDEV = as.double(hasdev),
            HASDF = as.double(hasdf),
            SCORE = as.double(score),
            RSS = as.double(rss),
            INFO = as.integer(info),
            LAMLIM = as.double(ctl$lamlim),
            MAXTBL = as.integer(ctl$maxtbl),
            NTBL = as.integer(ctl$ntbl),
            JOB = as.integer(ctl$job),
            TAU = as.double(ctl$tau),
            IWK = as.integer(iwk), 
            IWORK = as.integer(iwork), 
            YY = as.double(yy), 
            BASWK = as.double(baswk), 
            WKA = as.double(wka), 
            DES = as.double(des), 
            SIGMA = as.double(sigma), 
            ADIAG = as.double(adiag),
            SVALS = as.double(svals),
            TBL = as.double(tbl),
            WORK = as.double(work),
            VMU = as.integer(vmu),
            U = as.double(u),
            H = as.double(h),
            FITSWK = as.double(fitswk),
            Z = as.double(z),
            ZK = as.double(zk))[c("NB", "ISB", "COEF", 
            "HASFIT", "HASDEV", "HASDF", "SCORE", "RSS", "INFO","ADIAG","IDF")]

    # if (hasfit$INFO > 0) stop("GCVPACK/dsnsm error")

    idf <- hasfit$IDF
    isb <- hasfit$ISB
    isb <- isb[1:(length(isb)-1)]
    nb <- hasfit$NB
    selbas <- bas[,isb[1:nb]]
    A.diag <- hasfit$ADIAG	
	scr <- hasfit$SCORE
	scr <- scr[1:(length(scr)-1)]
	
    if (is.null(backward)) {
      if (ridge == T & nb > nnull) {
        coef <- hasfit$COEF
        fits <- hasfit$HASFIT
        sd <- hasfit$HASDEV
        df <- hasfit$HASDF
        A.diag <- hasfit$ADIAG  
      }
      if (ridge == F | nb == nnull) {
        fit1 <- glm(y~selbas-1)
        fits <- fit1$fitted.values
        coef <- fit1$coef
        sd <- sqrt(fit1$deviance/fit1$df.residual)
        df <- nb
        A.diag <- hat(selbas)
      }     
    }

    if (!is.null(backward) & nb == nnull) {
      fit1 <- glm(y~selbas-1)
      fits <- fit1$fitted.values
      coef <- fit1$coef
      sd <- sqrt(fit1$deviance/fit1$df.residual)
      df <- nb
      A.diag <- hat(selbas)	
    }
	                        
    if (!is.null(backward) & nb > nnull) {

      tmpbas <- bas[,isb[1:nb]]
      cstep <- regsubsets(tmpbas,y,force.in=c(1:nnull),method="backward",intercept=F)

      if (nb==maxbas) isb <- isb[cstep$vorder]
      if (nb<maxbas) isb <- isb[c(cstep$vorder,(nb+1):maxbas)]
      tmp <- cstep$rss[1:nb]
      if (backward == "GCV") scr <- tmp/nobs/(1-(1:nb)/nobs)**2
      if (backward == "AIC") scr <- nobs*log(tmp/nobs) + 2*(1:nb)
      if (backward == "BIC") scr <- nobs*log(tmp/nobs) + log(nobs)*(1:nb)
      if (backward == "Cp") scr <- tmp + 2*(1:nb)*tmp[nb]/(nobs-nb)
      tmp1 <- scr==min(scr)
      cum <- 0         
      cumsum <- 0
      while (cumsum==0) {
        cum <- cum+1
        cumsum <- sum(tmp1[1:cum])
      }
      nb <- cum
      selbas <- bas[,isb[1:nb]]
      if (ridge == F | nb == nnull) {
        fit1 <- glm(y~selbas-1)
        fits <- fit1$fitted.values
        coef <- fit1$coef
        sd <- sqrt(fit1$deviance/fit1$df.residual)
        df <- nb
      }
      if (ridge == T & nb > nnull) {
        pen.mat <- bdiag(list(matrix(0,nnull,nnull),selbas[1:(nb-nnull),(nnull+1):nb]+ctl$delta*diag(rep(1,nb-nnull))))
        fit1 <- snsm1(selbas, y, S=pen.mat)
        fits <- fit1$fitted.values
        pen.info <- fit1$penalty.info
        sd <- sqrt(pen.info[3]/(nobs-(nobs-pen.info[4]-nnull)*idf-nnull))
        df <- nobs - pen.info[4]
        coef <- fit1$coef
	  A.diag <- fit1$adiag
      }
    }

    # Identify the library index and basis index of the chosen bases
    chosen.bases.index <- chosen.bases.matrix <- NULL
    if (nb>nnull) {
    if (nnull==0) index.sel <- isb else index.sel <- isb[-(1:nnull)]-nnull
    L <- length(baseslist)
    liblen <- libnum <- NULL
    for (j in 2:L) {
      liblen <- c(liblen,ncol(as.matrix(baseslist[[j]])))
      libnum <- c(libnum,1:ncol(as.matrix(baseslist[[j]])))
    }
    libindex <- rep(2:L,liblen)
    lib.num <- libindex[index.sel]
    bas.num <- libnum[index.sel]  
    chosen.bases.full <- cbind(lib.num,bas.num)
	chosen.bases.trim <- chosen.bases.full[1:(nb-nnull),]
    chosen.bases.matrix <- selbas[,(nnull+1):ncol(selbas)]
    }
	if (!is.null(backward)) score <- scr[nnull:nb] else score <- scr[nnull:length(scr)]
	
	# Summary information
	residuals <- y-fits
	sigma_sq <- sum(residuals^2)/(nobs-idf*nb)
	dof <- idf*nb

    return(list(coefficients=coef,residuals=residuals,fitted.values=fits,chosen.bases.full=as.matrix(chosen.bases.full),chosen.bases.trim=as.matrix(chosen.bases.trim),chosen.bases.matrix=as.matrix(chosen.bases.matrix),score=score,gdf=NA,idf=idf,nb=nb,sigma_sq=sigma_sq,dof=dof,index.bas=isb))

}

    # Function for ridge regression
    snsm1 <- function(X, Y, method = 1, S = diag(ncol(X))) {
    # op=0 using GML; op= 1 using GCV; op=2 using UBR
    call <- match.call()
    n <- dim(S)[1]
    storage.mode(S) <- "double"
    rk.penalty.matrix <- .Fortran("dchdc", x = S , as.integer(n), as.integer(n), double(n), piv = as.integer(rep(0,n)), as.integer(1), rank = integer(1))$rank

    #print(rk.penalty.matrix)
        npar <- ncol(X)
        nobs <- nrow(X)
        tau <- 1
        lamlim <- c(-20, 20)
        ldtbl <- ntbl <- 1000
        tbl <- matrix(rep(0, ldtbl * 3), ncol = 3)
        auxtbl <- matrix(rep(0, 9), ncol = 3)
        nnull <- npar - rk.penalty.matrix
        lwa <- (npar - nnull) * (npar - 2 * nnull + 2 + nobs) + npar + nobs
        liwa <- 2 * npar - nnull
        job <- 1005    # tbl GCV data,
        z <- .Fortran("dsnsm1",
                X = as.double(X),
                LdX = as.integer(nrow(X)),
                Y = as.double(Y),
                S = as.double(S),
                LdS = as.integer(nrow(S)),
                nobs = as.integer(nobs),
                npar = as.integer(npar),
                nnull = as.integer(nnull),
                adiag = double(nobs),
                tau = as.double(tau),
                lamlim = as.double(lamlim),
                ntbl = as.integer(ntbl),
                penalty.info = double(5),
                iout = integer(3),
                coef = double(npar),
                svals = double(npar - nnull),
                gcv.info = as.double(tbl),
                Ldtbl = as.integer(ldtbl),
                auxtbl = as.double(auxtbl),
                iwork = integer(liwa),
                liwa = as.integer(liwa),
                work = double(lwa),
                lwa = as.integer(lwa),
                job = as.integer(job),
                info = integer(1),
                vm = as.integer(method))

        return(list(coefficients = z$coef, residuals = (Y - z$Y),
              fitted.values = z$Y, penalty.matrix = S,
              penalty.info = z$penalty.info, gcv.info = z$gcv.info,
              aux.info = z$aux.info, call = call,
              info = z$info, adiag = z$adiag))
    }

######################################################
# Basis Selection from Multiple Libraries - Combined #
######################################################


bsmlc <- function(y,baseslist,maxbas=30,sub.maxbas=c(),backward=NULL,
                  approach="gdf",
                  gdf.control=list(nrep=100,tauhat="Rice"),
                  gcvpk.control=list(delta=.0001,job=1005,lamlim=c(-6,2),tau=10,ntbl=100,maxtbl=100)) 
{
    n <- nobs <- length(y)
    starter <- ifelse(is.null(baseslist[[1]]),2,1)
	# Check if the number of rows of all libraries are the same
	for (i in starter:length(baseslist)) {
	lib.dim=nrow(as.matrix(baseslist[[i]]))
	if (lib.dim!=nobs) stop("Library ", i, " has the wrong number of rows.", "\n")
	}

	coefficients <- coef <- residuals <- res <- fitted.values <- fitted <- score <- score.null <- gdf <- gdf.new <- idf <- idf.new <- nb <- NULL
	chosen.bases.full <- chosen.bases.trim <- chosen.bases.matrix <- NULL
	gcv <- rep(0,maxbas)
	nrep <- gdf.control$nrep
	tauhat <- gdf.control$tauhat
	nnull <- ifelse(is.null(baseslist[[1]]),0,ncol(baseslist[[1]]))
       for (i in 2:length(baseslist)) colnames(baseslist[[i]]) <- as.character(1:ncol(baseslist[[i]]))

      # If a subset of bases of length sub.maxbas is requested for 
      # the libraries, then HAS is used to identify the best subsets of knots.

      if (!is.null(sub.maxbas)) {
	    orig.index <- matrix(length(sub.maxbas)*max(sub.maxbas),max(sub.maxbas),length(sub.maxbas))
        for (i in 1:length(sub.maxbas)) {
          if (sub.maxbas[i] < ncol(baseslist[[i+1]])) {
            has.obj <- has(y,list(baseslist[[1]],baseslist[[i+1]]),maxbas=nnull+sub.maxbas[i],gdf.control=gdf.control,gcvpk.control=gcvpk.control)
            baseslist[[i+1]] <- baseslist[[i+1]][,(has.obj$index.bas[(nnull+1):length(has.obj$index.bas)]-nnull)]
	     orig.index[,i]=as.numeric(colnames(baseslist[[i+1]]))
          }
        }
      }

      bas <- basel <- NULL
      for (i in 1:length(baseslist)) {
      bas <- cbind(bas,baseslist[[i]])
      }

	# Double-checking if total number of subknots chosen is greater than maxbas
    totalbases <- 0
    for (i in starter:length(baseslist)) totalbases <- totalbases+ncol(baseslist[[i]])
    if (totalbases<maxbas) stop("Total number of basis functions in all libraries is less than MAXBAS.","\n")  

    # Check if any two columns (bases) among the libraries are the same
    for (i in 2:length(baseslist)) basel <- cbind(basel,baseslist[[i]])
	numcol <- ncol(as.matrix(basel))
	coldist <- 0
	for (j in 1:(numcol-1)) {
		if (var(basel[,j])==0) warning("Column ", j, " is a constant.","\n")
		for (k in (j+1):numcol) {
			coldist <- sum((basel[,j]-basel[,k])^2)
			if (abs(coldist)<1e-20) warning("Column ", j, " and column ", k, " in the combined non-null libraries are the same.","\n","Fit is likely wrong.","\n")
		}
	}
	if (var(basel[,numcol])==0) warning("Column ", numcol, " is a constant.","\n")
	
#    if (min(diag(cov(basel)))>0) colcor <- cor(basel) else stop("Constant function in non-null libraries.","\n")
#    maxcor <- max(colcor-diag(rep(1,ncol(as.matrix(basel)))))
#    if (abs(maxcor-1)<1e-20) stop("Same basis has appeared twice among the libraries.","\n")

    # Compute Rice Estimator of standard deviation

    if (tauhat == "Rice") tauhat <- .5*sqrt(mean(diff(y)^2)/2)
	else if (tauhat == "Robust") {
	  gaps=abs(diff(y))
	  gaps.lb=quantile(gaps,0.025)
	  gaps.ub=quantile(gaps,0.975)
	  gaps=gaps*(1*(gaps>=gaps.lb & gaps<=gaps.ub))
	  tauhat <- .5*sqrt(mean((gaps[gaps>0])^2)/2)
	}
      else if (tauhat == "MAD") {
        tauhat <- 0.5*median(abs(diff(y)))
      }


    startpt <- ifelse(is.null(baseslist[[1]]),1,nnull)
    tmp <- bf.cost2(y,bas,nnull=nnull,maxbas=maxbas,wkamax=maxbas,nrep=nrep,tauhat=tauhat,approach=approach)
    gdf <- tmp$gdfhat
    gdf <- gdf[gdf<n]
    gdf.new <- gdf[(nnull+1):length(gdf)]
    idf.new <- tmp$idfhat[(nnull+1):length(gdf)]
    index.sel <- tmp$isb[(nnull+1):length(gdf)]
    rss <- tmp$rss[startpt:length(gdf)]
    gdf <- gdf[startpt:length(gdf)]
    if (approach=="gdf") gcv <- rss/(n-gdf)^2
    else if (approach=="covpen") gcv <- rss+log(n)*gdf
    gcv[is.na(gcv)] <- max(gcv[!is.na(gcv)])
    kk <- min((startpt:(length(gcv)+startpt-1))[gcv==min(gcv)])
    selbas <- bas[,tmp$isb[1:kk]]
    fit1 <- glm(y~selbas-1)
    fitted <- fit1$fitted
    coef <- fit1$coef
    res <- fit1$res

    # Identify the library index and basis index of the chosen bases
    begin <- ifelse(is.null(baseslist[[1]]),2,1)
    L <- length(baseslist)
    liblen <- libnum <- NULL
    for (j in begin:L) {
      liblen <- c(liblen,ncol(baseslist[[j]]))
      libnum <- c(libnum,1:ncol(baseslist[[j]]))
    }
    libindex <- rep(begin:L,liblen)
    lib.num <- libindex[index.sel]
    bas.num <- libnum[index.sel]  
    chosen.bases.full <- cbind(lib.num,bas.num)
    chosen.bases.trim <- chosen.bases.full[1:(kk-nnull),]
    if (!is.null(sub.maxbas)) {
      for (i in 1:length(sub.maxbas)) {
        libnum.full=chosen.bases.full[,1]
        temp.full=chosen.bases.full[libnum.full==(i+1),2]
        chosen.bases.full[libnum.full==(i+1),2]=orig.index[temp.full,i]
        libnum.trim=chosen.bases.trim[,1]
        temp.trim=chosen.bases.trim[libnum.trim==(i+1),2]
        chosen.bases.trim[libnum.trim==(i+1),2]=orig.index[temp.trim,i]
      }
    }
    chosen.bases.matrix <- selbas[,(nnull+1):kk]
    nb=kk
	score=gcv
	
    # Backward Elimination (output of GDFs and IDF will not be affected)
    if (!is.null(backward) && !is.null(chosen.bases.full)) {
    if (nnull==0) importance=regsubsets(selbas,y,method="backward",intercept=F)
    else importance=regsubsets(selbas,y,force.in=c(1:nnull),method="backward",intercept=F)

    imp.rss <- importance$rss[nnull:kk]
    gdf.trim <- gdf[1:kk]
    gdf.trim <- gdf.trim[importance$vorder[1:kk]]
	gdf.trim <- gdf.trim[nnull:kk]
  
    if (backward == "GCV") scr <- imp.rss/(nobs-(gdf.trim))^2
    if (backward == "AIC") scr <- nobs*log(imp.rss/nobs) + 2*(gdf.trim)
    if (backward == "BIC") scr <- nobs*log(imp.rss/nobs) + log(nobs)*(gdf.trim)
    if (backward == "Cp")  scr <- imp.rss + 2*(gdf.trim)*imp.rss[kk-nnull]/(nobs-(gdf.trim))

    trim.again <- order(scr)[1]-1
	chosen.bases.trim.neworder <- chosen.bases.trim[(importance$vorder[(nnull+1):kk]-nnull),]
    chosen.bases.trim <- chosen.bases.trim[(importance$vorder[(nnull+1):kk]-nnull),][1:trim.again,]
	# The following line controls whether to preserve the original bases order
	chosen.bases.full <- rbind(chosen.bases.trim.neworder,chosen.bases.full[-(1:(nb-nnull)),])
    chosen.bases.matrix <- chosen.bases.matrix[,(importance$vorder[(nnull+1):kk]-nnull)][,1:trim.again]
    final.bases.backward <- cbind(baseslist[[1]],chosen.bases.matrix)
    fit <- glm(y~final.bases.backward-1)
    fitted <- fit$fitted
    coef <- fit$coef
    res <- fit$res
    nb=trim.again+nnull
	score=scr
    }
	
	# Summary information
	gen_deg_fre <- ifelse(is.null(baseslist[[1]]),gdf.new[nb],gdf.new[(nb-nnull)])
	sigma_sq <- sum(res^2)/(nobs-gen_deg_fre)

    return(list(coefficients=coef,residuals=res,fitted.values=fitted,chosen.bases.full=as.matrix(chosen.bases.full),chosen.bases.trim=as.matrix(chosen.bases.trim),chosen.bases.matrix=as.matrix(chosen.bases.matrix),score=score,gdf=gdf.new,idf=idf.new,nb=nb,sigma_sq=sigma_sq,dof=gen_deg_fre))
}

######################################################
# Basis Selection from Multiple Libraries - Separate #
######################################################


bsmls <- function(y,baseslist,maxbas=30,sub.maxbas=c(),backward=NULL,
                  gdf.control=list(nrep=100,tauhat="Rice"),
                  gcvpk.control=list(delta=.0001,job=1005,lamlim=c(-6,2),tau=10,ntbl=100,maxtbl=100)) 
{
  nobs=length(y)
  starter <- ifelse(is.null(baseslist[[1]]),2,1)
  # Check if the number of rows of all libraries are the same
  for (i in starter:length(baseslist)) {
    lib.dim=nrow(as.matrix(baseslist[[i]]))
    if (lib.dim!=nobs) stop("Library ", i, " has the wrong number of rows.", "\n")
  }

	coefficients <- coef <- residuals <- res <- fitted.values <- fitted <- score <- score.null <- gdf <- idf <- nb <- NULL
  chosen.bases.full <- chosen.bases.trim <- chosen.bases.matrix <- NULL
  bas <- basel <- NULL
  n=length(y)
  nrep=gdf.control$nrep
  tauhat=gdf.control$tauhat
  nnull=ifelse(is.null(baseslist[[1]]),0,ncol(baseslist[[1]]))
  updatevar=0
  for (i in 2:length(baseslist)) colnames(baseslist[[i]]) <- as.character(1:ncol(baseslist[[i]]))



  # If a subset of bases of length sub.maxbas is requested for 
  # the libraries, then HAS is used to identify the best subsets of knots.

  if (!is.null(sub.maxbas)) {
	orig.index <- matrix(length(sub.maxbas)*max(sub.maxbas),max(sub.maxbas),length(sub.maxbas))
    for (i in 1:length(sub.maxbas)) {
      if (sub.maxbas[i] < ncol(baseslist[[i+1]])) {
        has.obj <- has(y,list(baseslist[[1]],baseslist[[i+1]]),maxbas=nnull+sub.maxbas[i],gdf.control=gdf.control,gcvpk.control=gcvpk.control)
        baseslist[[i+1]] <- baseslist[[i+1]][,(has.obj$index.bas[(nnull+1):length(has.obj$index.bas)]-nnull)]
	 orig.index[,i]=as.numeric(colnames(baseslist[[i+1]]))
      }
    }
  }


  for (i in 1:length(baseslist)) {
  bas <- cbind(bas,baseslist[[i]])
  }

	# Double-checking if total number of subknots chosen is greater than maxbas
  totalbases <- 0
  for (i in starter:length(baseslist)) totalbases <- totalbases+ncol(baseslist[[i]])
  if (totalbases<maxbas) stop("Total number of basis functions in all libraries is less than MAXBAS.","\n")  

    # Check if any two columns (bases) among the libraries are the same
    for (i in 2:length(baseslist)) basel <- cbind(basel,baseslist[[i]])
	numcol <- ncol(as.matrix(basel))
	coldist <- 0
	for (j in 1:(numcol-1)) {
		if (var(basel[,j])==0) warning("Column ", j, " is a constant.","\n")
		for (k in (j+1):numcol) {
			coldist <- sum((basel[,j]-basel[,k])^2)
			if (abs(coldist)<1e-20) warning("Column ", j, " and column ", k, " in the combined non-null libraries are the same.","\n","Fit is likely wrong.","\n")
		}
	}
	if (var(basel[,numcol])==0) warning("Column ", numcol, " is a constant.","\n")
	
#    if (min(diag(cov(basel)))>0) colcor <- cor(basel) else stop("Constant function in non-null libraries.","\n")
#    maxcor <- max(colcor-diag(rep(1,ncol(as.matrix(basel)))))
#    if (abs(maxcor-1)<1e-20) stop("Same basis has appeared twice among the libraries.","\n")
  

  #### Rice Estimator of the variance if sigma^2 not supplied ####
  if (tauhat == "Rice") {
  est.var=mean(diff(y)^2)/2
  s.hat=sqrt(est.var)
  }
  if (tauhat == "Robust") {
    gaps=abs(diff(y))
    gaps.lb=quantile(gaps,0.025)
    gaps.ub=quantile(gaps,0.975)
    gaps=gaps*(1*(gaps>=gaps.lb & gaps<=gaps.ub))
    s.hat <- sqrt(mean((gaps[gaps>0])^2)/2)
    est.var=s.hat^2
  }
  if (tauhat == "MAD") {
    s.hat=median(abs(diff(y)))
    est.var=s.hat^2
  }
  if (is.numeric(tauhat)) {
  est.var=4*tauhat^2
  s.hat=sqrt(est.var)
  }

  rss.sel<-bas.gdf<-rep(0,maxbas)
  lib=baseslist
  new.sel=matrix(0,length(y),length(baseslist))
  lib.cost<-cum.bas.cost<-rss<-bas.cost<-lib.size<-dpc<-index.new<-rep(0,length(baseslist))
  lib.index=list()
  for (j in starter:length(baseslist)) {
  lib.size[j]=ncol(as.matrix(baseslist[[j]]))
  lib.index[[j]]=1:ncol(as.matrix(baseslist[[j]]))
  }
  lib.var<-new.source<-sel.bases<-NULL
  lib.num<-bas.num<-rep(0,maxbas)
  chosen.bases.index=cbind(lib.num,bas.num)
  df.var<-gdf<-0
  score.null=NA
  chosen.bases.full=NULL
  wka.tmp <- wkall.tmp <- yy.tmp <- yyptb.tmp <- NULL

  # Pre-evaluation of library cost for each library for up to "maxbas" number of bases selected
  gdf.pre<-idf.pre<-matrix(0,maxbas,length(baseslist)-1)
  for (w in 2:length(baseslist)) {
    pre.eval=bf.cost2(y,cbind(baseslist[[1]],baseslist[[w]]),nnull=nnull,maxbas=min(maxbas,nnull+ncol(baseslist[[w]])),wkamax=maxbas,nrep=nrep,tauhat=0.5*s.hat)
    gdf.pre[,w-1][1:length(pre.eval$gdf)]=pre.eval$gdf
    idf.pre[,w-1][1:length(pre.eval$idf)]=pre.eval$idf
  }

  if (nnull!=0) {
  init.eval=bf.cost2(y,baseslist[[1]],nnull=nnull,maxbas=nnull,wkamax=maxbas,nrep=nrep,tauhat=0.5*s.hat)
  wka=init.eval$wka
  wkall=init.eval$wkall
  ptb=init.eval$ptb
  yy=init.eval$yy
  yyptb=init.eval$yyptb
  }

  if (nnull==0) {
  init.eval=bf.cost2(y,bas,nnull=nnull,maxbas=1,wkamax=maxbas,nrep=nrep,tauhat=0.5*s.hat)
  wka=init.eval$wka
  wkall=init.eval$wkall
  ptb=init.eval$ptb
  yy=init.eval$yy
  yyptb=init.eval$yyptb
  begin <- 2
  L <- length(baseslist)
  liblen <- libnum <- NULL
  for (j in begin:L) {
    liblen <- c(liblen,ncol(baseslist[[j]]))
    libnum <- c(libnum,1:ncol(baseslist[[j]]))
  }
  libindex <- rep(begin:L,liblen)
  index.sel <- init.eval$isb
  pick <- lib.num <- libindex[index.sel]
  index.new[pick] <- bas.num <- libnum[index.sel]  
  sel.bases=cbind(sel.bases,as.matrix(lib[[pick]])[,index.new[pick]])
  chosen.bases.index[1,]=c(pick,lib.index[[pick]][index.new[pick]])
  bas.gdf[1]=init.eval$gdf
  rss.sel[1]=init.eval$rss
#  gdf=bas.gdf[i]
  lib.size[pick]=0
  if (ncol(lib[[pick]])>1) {
  lib[[pick]]=as.matrix(lib[[pick]])[,-index.new[pick]]
  lib.index[[pick]]=lib.index[[pick]][-index.new[pick]]
  lib.size[pick]=ncol(lib[[pick]])
  }
  }

  # Choose a candidate basis from each library and compute the corresponding basis cost and library cost
  i=ifelse(is.null(baseslist[[1]]),2,1)
  while ((i <= (maxbas-nnull)) && (gdf < length(y))) {
  # Calculate costs for bases from different libraries
  for (j in 2:length(baseslist)) {
    dpc[j]=Inf
    if (lib.size[j] > 0) {
	new.source=cbind(baseslist[[1]],sel.bases,lib[[j]])
	bas.eval=pick.bas(yy,yyptb,new.source,wka=wka,wkall=wkall,wkamax=maxbas,ptb=ptb,nnull=nnull+i-1,nrep=nrep,tauhat=0.5*s.hat)
	wka.tmp[[j]]=bas.eval$wka
	wkall.tmp[[j]]=bas.eval$wkall
	yy.tmp[[j]]=bas.eval$yy
	yyptb.tmp[[j]]=bas.eval$yyptb
#	bas.cost[j]=bas.eval$idfhat
	cum.bas.cost[j]=bas.eval$gdfhat
	bas.cost[j]=max(cum.bas.cost[j]-(i-1),1)
	rss[j]=bas.eval$rss
	lib.cost[j]=idf.pre[nnull+ncol(baseslist[[j]])-lib.size[j]+1,j-1]
	new.sel[,j]=new.source[,bas.eval$isb]
       index.new[j]=bas.eval$isb-(nnull+i-1)
#	status=c(j,index.new[j],bas.cost[j])
#	print(status)
    }
  }

  # Update variance estimate
  if (updatevar==1) {
  lib.var=cbind(baseslist[[1]],sel.bases,new.sel[,-1])
  df.var=nnull+i-1+sum(bas.cost)
  if (df.var>=length(y)) stop("Effective degree of freedom exceeds sample size.","\n")
  lm.var=lm(y~lib.var)
  est.var=sum((y-lm.var$fit)^2/(length(y)-df.var))
  s.hat=sqrt(est.var)
  }

  # Compute the Doubly Penalized Criterion (DPC)
  for (j in 2:length(baseslist)) {
    dpc[j]=rss[j]+est.var*log(length(y))*(cum.bas.cost[j])+est.var*lib.cost[j]
  }

  # Compare bases from different libraries based on DPC and make a choice
  dpc2=dpc[2:length(dpc)]
  pick=order(dpc2)[1]+1
  sel.bases=cbind(sel.bases,cbind(lib[[pick]])[,index.new[pick]])

pick.eval=pick.bas(yy,yyptb,cbind(baseslist[[1]],sel.bases),wka=wka,wkall=wkall,wkamax=maxbas,ptb=ptb,nnull=nnull+i-1,nrep=nrep,tauhat=0.5*s.hat)

  wka=wka.tmp[[pick]]
  wkall=pick.eval$wkall
#  wkall=wkall.tmp[[pick]]
  yy=yy.tmp[[pick]]
  yyptb=pick.eval$yyptb
#  yyptb=yyptb.tmp[[pick]]
  chosen.bases.index[i,]=c(pick,lib.index[[pick]][index.new[pick]])
  rss.sel[i]=rss[pick]

  # GDF calculation #
  bas.gdf[i]=ifelse(i==1,(bas.cost[pick]+nnull),bas.cost[pick])
  if (i>1) bas.gdf[i]=bas.gdf[i]+bas.gdf[i-1]
  gdf=bas.gdf[i]  

  lib.size[pick]=0
  if (ncol(as.matrix(lib[[pick]]))>1) {
  lib[[pick]]=lib[[pick]][,-index.new[pick]]
  lib.index[[pick]]=lib.index[[pick]][-index.new[pick]]
  lib.size[pick]=ncol(as.matrix(lib[[pick]]))
  }

  i=i+1
  }
  maxbas.sel=length(bas.gdf[1:(i-1)])

  # Trimming back
  if (!is.null(lib[[1]])) {
    fit.null=glm(y~lib[[1]]-1)
    score.null=(fit.null$dev)/(nobs-ncol(lib[[1]]))^2
    fitted=fit.null$fitted
    coef=fit.null$coef
    res=fit.null$res
    nb=nnull
  }

  score=rep(0,maxbas.sel)
  for (i in 1:maxbas.sel) score[i]=rss.sel[i]/(length(y)-bas.gdf[i])^2
  trim=order(score)[1]
  if (is.na(score.null) || (!is.na(score.null)&&(score.null>score[trim]))) {
  final.bases=cbind(baseslist[[1]],sel.bases[,1:trim])
  fit=glm(y~final.bases-1)
  fitted=fit$fitted
  coef=fit$coef
  res=fit$res
  gdf=bas.gdf[1:maxbas.sel]
  idf=(bas.gdf-nnull)/(maxbas.sel-nnull)
  idf=idf[1:maxbas.sel]
  idf[idf<0]=0
  chosen.bases.full=chosen.bases.index[1:maxbas.sel,]
  chosen.bases.trim=chosen.bases.index[1:trim,]
  if (!is.null(sub.maxbas)) {
    for (i in 1:length(sub.maxbas)) {
      libnum.full=chosen.bases.full[,1]
      temp.full=chosen.bases.full[libnum.full==(i+1),2]
      chosen.bases.full[libnum.full==(i+1),2]=orig.index[temp.full,i]
      libnum.trim=chosen.bases.trim[,1]
      temp.trim=chosen.bases.trim[libnum.trim==(i+1),2]
      chosen.bases.trim[libnum.trim==(i+1),2]=orig.index[temp.trim,i]
    }
  }
  chosen.bases.matrix=sel.bases[,1:trim]
  nb=trim+nnull
  }
  score=c(score.null,score)
  
  # Backward Elimination (output of GDFs and IDFs will not be affected)
  if (!is.null(backward) && !is.null(chosen.bases.full)) {
    if (nnull==0) importance=regsubsets(final.bases,y,method="backward",intercept=F)
    else importance=regsubsets(final.bases,y,force.in=c(1:nnull),method="backward",intercept=F)

    imp.rss=importance$rss[nnull:(nnull+trim)]
    gdf.trim=bas.gdf[1:(nnull+trim)]
    gdf.trim=gdf.trim[importance$vorder[1:(nnull+trim)]]
	gdf.trim=gdf.trim[nnull:(nnull+trim)]
  
    if (backward == "GCV") scr <- imp.rss/(nobs-(gdf.trim))^2
    if (backward == "AIC") scr <- nobs*log(imp.rss/nobs) + 2*(gdf.trim)
    if (backward == "BIC") scr <- nobs*log(imp.rss/nobs) + log(nobs)*(gdf.trim)
    if (backward == "Cp")  scr <- imp.rss + 2*(gdf.trim)*imp.rss[trim]/(nobs-(gdf.trim))

    trim.again=order(scr)[1]-1
	chosen.bases.trim.neworder=chosen.bases.trim[(importance$vorder[(nnull+1):(nnull+trim)]-nnull),]
    chosen.bases.trim=chosen.bases.trim[(importance$vorder[(nnull+1):(nnull+trim)]-nnull),][1:trim.again,]
	# The following line controls whether to preserve the original bases order
	chosen.bases.full <- rbind(chosen.bases.trim.neworder,chosen.bases.full[-(1:(nb-nnull)),])
    chosen.bases.matrix=chosen.bases.matrix[,(importance$vorder[(nnull+1):(nnull+trim)]-nnull)][,1:trim.again]
    final.bases.backward=cbind(baseslist[[1]],chosen.bases.matrix)
    fit=glm(y~final.bases.backward-1)
    fitted=fit$fitted
    coef=fit$coef
	res=fit$res
    nb=trim.again+nnull
	score=scr
  }

  # Summary information
  gen_deg_fre <- ifelse(is.null(baseslist[[1]]),gdf[nb],gdf[(nb-nnull)])
  sigma_sq <- sum(res^2)/(n-gen_deg_fre)

  return(list(coefficients=coef,residuals=res,fitted.values=fitted,chosen.bases.full=as.matrix(chosen.bases.full),chosen.bases.trim=as.matrix(chosen.bases.trim),chosen.bases.matrix=as.matrix(chosen.bases.matrix),score=score,gdf=gdf,idf=idf,nb=nb,sigma_sq=sigma_sq,dof=gen_deg_fre))
}

#########################
# Basis Function Cost 2 #
# Updated Oct 13, 2009  #
#########################

# Selects up to maxbas number of basis functions and calculates the corresponding
# GDF's and IDF's. Two approaches are available: the one described in Ye's paper
# or the covariance penalty approach.
# Calls either LIBCOST2 (Ye) or LIBCOST3 (Covariance penalty)


bf.cost2 <- function(y,bas,nnull=NA,maxbas=NA,wkamax=NA,nrep=100,tauhat=NA,approach="gdf")
{
    nobs <- length(y)
    nbas <- ncol(bas)
 
    idfhat <- gdfhat <- rep(0,maxbas)
    if (nbas < maxbas) stop("nbas < maxbas")	
    isb <- rss <- seq(0, maxbas)
    yy <- y
    iwk <- rep(1, nbas)
    baswk <- bas 
    wka <- matrix(0,nobs,wkamax)
    wkall <- array(rep(0,nobs*wkamax*nrep),dim=c(nobs,wkamax,nrep))
    icb <- matrix(0,maxbas,nrep)
    ptb <- matrix(0,nobs,nrep) 
    u <- h <- z <- zk <- rep(0,nobs)
    fits <- array(rep(0,nobs*maxbas*nrep),dim=c(nobs,maxbas,nrep))
    fitswk <- matrix(0,nobs,maxbas)
    yyptb <- matrix(0,nobs,nrep)
    if (approach=="gdf") routine="libcost2"
    else if (approach=="covpen") routine="libcost3"

    hasfit <- .Fortran(routine,
            NOBS = as.integer(nobs),
            Y = as.double(y),
            BAS = as.double(bas),
            NNULL = as.integer(nnull),
            NBAS = as.integer(nbas),
            MAXBAS = as.integer(maxbas),
            IDFHAT = as.double(idfhat),
	      nrep = as.integer(nrep),
	      GDFHAT = as.double(gdfhat),
            TAUHAT = as.double(tauhat),
            ISB = as.integer(isb),
            RSS = as.double(rss),
            IWK = as.integer(iwk), 
            YY = as.double(yy), 
            BASWK = as.double(baswk), 
            WKA = as.double(wka), 
		WKALL = as.double(wkall),
		WKAMAX = as.integer(wkamax),
		ICB = as.double(icb),
		PTB = as.double(ptb),
            U = as.double(u),
            H = as.double(h),
            FITS = as.double(fits),
            FITSWK = as.double(fitswk),
            Z = as.double(z),
            ZK = as.double(zk),
		YYPTB = as.double(yyptb)
	      )[c("IDFHAT","GDFHAT","ISB","IWK","RSS","ICB","PTB","FITS","U","H","Z","ZK","WKA","WKALL","BASWK","YY","FITSWK","YYPTB")]
    rss <- hasfit$RSS[1:maxbas]
    isb <- hasfit$ISB[1:maxbas]
    iwk <- hasfit$IWK
    idfhat <- hasfit$IDFHAT 
    gdfhat <- hasfit$GDFHAT
    icb <- matrix(as.integer(hasfit$ICB),ncol=nrep,byrow=F)
    ptb <- matrix(hasfit$PTB,ncol=nrep,byrow=F)
    fits <- array(hasfit$FITS,dim=c(nobs,maxbas,nrep))
    u=hasfit$U
    h=hasfit$H
    z=hasfit$Z
    zk=hasfit$ZK
    wka=matrix(hasfit$WKA,nrow=nobs)
    wkall=array(hasfit$WKALL,dim=c(nobs,wkamax,nrep))
    basnew=matrix(hasfit$BASWK,nrow=nobs)
    yy=hasfit$YY
    yyptb=matrix(hasfit$YYPTB,nrow=nobs)
    last.fit=matrix(hasfit$FITSWK,nobs)

    return(list(isb=isb,iwk=iwk,rss=rss,gdfhat=gdfhat,icb=icb,ptb=ptb,idfhat=idfhat,fits=fits,u=u,h=h,z=z,zk=zk,wka=wka,wkall=wkall,basnew=basnew,yy=yy,yyptb=yyptb,last.fit=last.fit))
}

################
# Pick Basis   #
# Oct 13, 2009 #
################

# Determines cost of adding a basis from a library given that a set of bases
# have already been selected for the model.
# Calls Fortran subroutine PICKBAS.


pick.bas <- function(y,yyptb,bas,wka,wkall,wkamax,ptb,nnull,nrep,tauhat)
{
    nobs <- length(y)
    nbas <- ncol(bas)
    idfhat <- gdfhat <- 0
    isb <- rss <- 0
    yy <- y
    baswk <- bas 
    u <- h <- z <- zk <- rep(0,nobs)
    fits <- matrix(0,nobs,nrep)
    fitswk <- rep(0,nobs)

    hasfit <- .Fortran("PICKBAS",
            NOBS = as.integer(nobs),
            Y = as.double(y),
            BAS = as.double(bas),
            NNULL = as.integer(nnull),
            NBAS = as.integer(nbas),
            IDFHAT = as.double(idfhat),
	      nrep = as.integer(nrep),
	      GDFHAT = as.double(gdfhat),
            TAUHAT = as.double(tauhat),
            ISB = as.integer(isb),
            RSS = as.double(rss),
            YY = as.double(yy), 
            BASWK = as.double(baswk), 
            WKA = as.double(wka), 
		WKALL = as.double(wkall),
		WKAMAX = as.integer(wkamax),
		PTB = as.double(ptb),
            U = as.double(u),
            H = as.double(h),
            FITS = as.double(fits),
            FITSWK = as.double(fitswk),
            Z = as.double(z),
            ZK = as.double(zk),
		YYPTB = as.double(yyptb)
	      )[c("IDFHAT","GDFHAT","ISB","RSS","FITS","U","H","Z","ZK","WKA","WKALL","BASWK","YY","YYPTB","FITSWK")]

    rss <- hasfit$RSS
    isb <- hasfit$ISB
    idfhat <- hasfit$IDFHAT 
    gdfhat <- hasfit$GDFHAT
    fits <- matrix(hasfit$FITS,nobs)
    u=hasfit$U
    h=hasfit$H
    z=hasfit$Z
    zk=hasfit$ZK
    wka=matrix(hasfit$WKA,nobs)
    wkall=array(hasfit$WKALL,dim=c(nobs,wkamax,nrep))
    basnew=matrix(hasfit$BASWK,nobs)
    yy=hasfit$YY
    yyptb=matrix(hasfit$YYPTB,nobs)
    last.fit=hasfit$FITSWK

    return(list(isb=isb,rss=rss,gdfhat=gdfhat,idfhat=idfhat,fits=fits,u=u,h=h,z=z,zk=zk,wka=wka,wkall=wkall,wkamax=wkamax,basnew=basnew,yy=yy,yyptb=yyptb,last.fit=last.fit))
}
